#!/bin/bash
#==============================================================================
# Copyright and license info is available in the LICENSE file included with
# the Server Deployment Package (SDP), and also available online:
# https://swarm.workshop.perforce.com/projects/perforce-software-sdp/view/main/LICENSE
#------------------------------------------------------------------------------
set -u

#==============================================================================
# Declarations and Environment

declare ThisScript="${0##*/}"
declare SDPInstance=UnsetSDPInstance
declare Version=4.1.0
declare -i NoOp=0
declare -i ErrorCount=0
declare -i WarningCount=0
declare -i PreflightOnly=0
declare -i DoProtectionsCommentStyleUpgrade=0
declare -i DoP4DMajorVersionUpgrade=0
declare -i P4D2019Plus=0
declare -i SilentMode=0
declare UpgradeJournal=
declare ExesDir=/p4/sdp/exes
declare ExeList="p4 p4d p4broker p4p"
declare UpgradeExeList=
declare UpgradeServiceList=
declare Exe=
declare ExePath=
declare NewVersionNamedExe=
declare InstanceLinkPath=
declare VersionLinkPath=
declare OldMajorVersion=
declare OldBuildCL=
declare OldVersion=
declare NewMajorVersion=
declare NewBuildCL=
declare NewVersion=
declare Log=

#==============================================================================
# Local Functions
function msg () { echo -e "$*"; }
function dbg () { [[ "$Debug" -eq 1 ]] && msg "DEBUG: $*"; }
function errmsg () { msg "\\nError: (line: ${BASH_LINENO[0]}) $*" >&2; ErrorCount+=1; }
function warnmsg () { msg "\\nWarning: (line: ${BASH_LINENO[0]}) $*"; WarningCount+=1; }
function bail() { errmsg "(line: ${BASH_LINENO[0]}) ${1:-Unknown Error}\\n" >&2; exit "${2:-1}"; }

#------------------------------------------------------------------------------
# This function takes as input an SDP version string, and returns a version
# id of the form YYYY.N.CL, where YYYY is the year, N is an incrementing
# release ID with a given year, and CL is a changelist identifier. The
# YYYY.N together comprise the major version, often shortened to YY.N, e.g.
# r20.1 for the 2020.1 release.
#
# The full SDP Version string looks something like this:
# Rev. SDP/MultiArch/2019.3/26494 (2020/04/23).
#
# This function parses that full string and returns a value like: 2019.3.26494
function get_sdp_version_from_string () {
   local versionString="${1:-}"
   local version=
   version="20${versionString##*/20}"
   version="${version%% *}"
   version="${version/\//.}"
   echo "$version"
}

#------------------------------------------------------------------------------
# Function: run ($cmd, $desc, $honorNoOp)
#
# Input:
#
# $cmd  - Command to execute, including arguments.?
#
# $desc - Description of command to run. Optional; default is no description.
#
# $honorNoOp - If set to 1, command is always executed; NoOp setting is ignored.
#    Optional, default is 0.
#------------------------------------------------------------------------------
function run () { 
   declare cmd="${1:-echo}"
   declare desc="${2:-}"
   declare honorNoOp="${3:-1}"

   [[ -n "$desc" ]] && msg "$desc"
   msg "Running: $cmd"

   if [[ "$NoOp" -eq 1 && "$honorNoOp" -eq 1 ]]; then
      msg "NO_OP: Would run $cmd"
      return 0
   else
      if $cmd; then
         return 0
      else
         return 1
      fi
   fi
}

#------------------------------------------------------------------------------
# Function: usage (required function)
#
# Input:
# $1 - style, either -h (for short form) or -man (for man-page like format).
# The default is -h.
#
# $2 - error message (optional).  Specify this if usage() is called due to
# user error, in which case the given message displayed first, followed by the
# standard usage message (short or long depending on $1).  If displaying an
# errror, usually $1 should be -h so that the longer usage message doesn't
# obsure the error message.
#
# Sample Usage:
# usage
# usage -h
# usage -man
# usage -h "Incorrect command line usage."
#------------------------------------------------------------------------------
function usage
{
   declare style=${1:--h}
   declare usageErrorMessage=${2:-Unset}

   if [[ "$usageErrorMessage" != Unset ]]; then
      echo -e "\\n\\nUsage Error:\\n\\n$usageErrorMessage\\n\\n"
   fi

   # tag::includeManual[]
   echo "USAGE for $ThisScript v$Version:

$ThisScript <instance> [-c] [-p] [-n] [-L <log>] [-d|-D]

or

$ThisScript [-h|-man]
"
   if [[ $style == -man ]]; then
      echo -e "
DESCRIPTION:

This script upgrades the following Helix Core software:

* p4d, the Perforce Helix Core server
* p4broker, the Helix Broker server
* p4p, the Helix Proxy server
* p4, the command line client

Details of each upgrade are described below. Prior to executing any upgrades, a
preflight check is done to help ensure upgrades will go smoothly.  Also, checks
are done to determine what (if any) of the above software products need to be
updated.

To prepare for an upgrade, new executables must be placed in the /p4/sdp/exes
directory.  Executables in this directory are not referenced by live running
servers, and so it is safe to ugprade files in this directory to stage for a
future upgrade at any time. The SDP PATH does not include this directory.

Upgrading p4d, the Perforce Helix Core Server

Upgrading p4broker and p4p (P4Proxy)

Upgrading p4, the command line client

	If your server can reach the Perforce FTP server over the public
	internet, this script can be used executed from the /p4/sdp/exes
	directory to get the latest executables.
	
	/p4/sdp/exes/get_latest_exes.sh

OPTIONS:
 <instance>
	Specify the SDP instance name to add.  This is a reference to the Perforce
	Helix Core data set. This defaults to the current instance based on the
	\$SDP_INSTANCE shell environment variable. If the SDP shell environment is
	not loaded, this option is required.

 -c	Specify '-c' to execute a command to upgrade the Protections table comment
	format after the p4d upgrade, by using a command like:

        p4 protect --convert-p4admin-comments -o | p4 -s protect -i

	By default, this Protections table conversion is not performed.  In some
	environments with custom policies related to update of the protections
	table, this command may not work.

 -p	Specify '-p' to halt processing after preflight checks are complete,
	and before actual processing starts. By default, procesing starts
	immediately upon successful completion of prelfight checks.

 -L <log>
	Specify the path to a log file, or the special value 'off' to disable
	logging.  By default, all output (stdout and stderr) goes to this file
	in the /p4/N/logs directory (where N is the SDP instance name):

	upgrade.p4_N.<datestamp>.log

	NOTE: This script is self-logging.  That is, output displayed on the screen
	is simultaneously captured in the log file.  Do not run this script with
	redirection operators like '> log' or '2>&1', and do not use 'tee.'

	NOTE: Logging can only be disabled with '-L off' if the '-n' or '-p' flags
	are used. Disabling logging for actual upgrades is not allowed.

DEBUGGING OPTIONS:
 -n	No-Op.  In No-Op mode, no actions that affect data or structures are taken.
	Instead, commands that would be run are displayed.

 -d     Increase verbosity for debugging.

 -D     Set extreme debugging verbosity, using bash '-x' mode. Also implies -d.

HELP OPTIONS:
 -h	Display short help message
 -man	Display man-style help message

EXAMPLES:
	Typical usage is with just the SDP instance name as an argument, e.g.
	instance '1' in this example:

	$ cd /p4/common/bin
	$ ./upgrade.sh 1

	This executes the upgrade after successful completion of preflight checks,
	and aborts if preflight checks detected any issues.  Alternately, to see
	if an upgrade is needed without starting, use the '-p' flag to execute only
	the preflight checks, as in this example:

	$ cd /p4/common/bin
	$ ./upgrade.sh 1 -p
"
   fi

   exit 1
}

#------------------------------------------------------------------------------
# Function: terminate
function terminate
{
   # Disable signal trapping.
   trap - EXIT SIGINT SIGTERM

   # Ensure a non-zero exit code if the trap was triggered by anything other
   # than the 'exit 0' at the end of the Main Program.
   [[ "$BASH_COMMAND" == *"exit 0"* ]] || ErrorCount+=1

   msg "\\n$ThisScript: EXITCODE: $ErrorCount\\n"
   [[ "$Log" != "off" ]] && msg "Log is: $Log"

   # With the trap removed, exit.
   exit "$ErrorCount"
}

#==============================================================================
# Command Line Processing

declare -i shiftArgs=0

set +u
while [[ $# -gt 0 ]]; do
   case $1 in
      (-c) DoProtectionsCommentStyleUpgrade=1;;
      (-p) PreflightOnly=1;;
      (-h) usage -h;;
      (-man) usage -man;;
      (-n) NoOp=1;;
      (-d) Debug=1;;
      (-si) SilentMode=1;;
      (-L) Log="$2"; shiftArgs=1;;
      (-D) Debug=1; set -x;; # Debug; use bash 'set -x' high verbosity debugging mode.
      (*)
         if [[ "$SDPInstance" == "UnsetSDPInstance" ]]; then
            SDPInstance="$1"
	 else
            usage -h "Instance $1 specified after already being specified as $SDPInstance. Only one instance can be specified."
	 fi
      ;;
   esac

   # Shift (modify $#) the appropriate number of times.
   shift; while [[ $shiftArgs -gt 0 ]]; do
      [[ $# -eq 0 ]] && usage -h "Incorrect number of arguments."
      shiftArgs=$shiftArgs-1
      shift
   done
done
set -u

# Default SDPInstance to the SDP_INSTANCE shell environment variable
# if no value was specified on the command line.
if [[ "$SDPInstance" == "UnsetSDPInstance" ]]; then
   if [[ -n "${SDP_INSTANCE:-}" ]]; then
      SDPInstance="$SDP_INSTANCE"
   else
      usage -h "No instance specified and SDP shell environment is not loaded."
   fi
fi

if [[ "$Log" == "off" ]]; then
   if [[ "$NoOp" -eq 0 && "$PreflightOnly" -eq 0 ]]; then
      usage -h "Disabling logging with '-L off' is only allowed with '-p' and/or '-n' flags."
   fi
fi

[[ "$SilentMode" -eq 1 && "$Log" == "off" ]] && \
   usage -h "Cannot use '-si' with '-L off'."

#==============================================================================
# Main Program

# Pre-Preflight checks occur before logging starts, to ensure essential utils
# (including those needed for logging) are availble in the PATH.
for util in awk date id sed tee touch; do
   if ! command -v $util > /dev/null; then
      errmsg "Cannot find this essential util in PATH: $util"
   fi
done

if [[ "$ErrorCount" -ne 0 ]]; then
   bail "Aborting due to essential tools not available in directories listed in the PATH."
fi

[[ -n "$Log" ]] ||\
   Log="${LOGS:-/tmp}/upgrade.p4_${SDPInstance}.$(date +'%Y%m%d-%H%M%S').log"

# Set LOGFILE to be compatible with functions like log() in backup_functions.sh.
export LOGFILE="${Log}"

if [[ "$Log" != off ]]; then
   touch "$Log" || bail "Couldn't touch log file [$Log]."
fi

trap terminate EXIT SIGINT SIGTERM

# Redirect stdout and stderr to the log file.
if [[ "$Log" != "off" ]]; then
   # Redirect stdout and stderr to a log file.
   if [[ "$SilentMode" -eq 0 ]]; then
      exec > >(tee "$Log")
      exec 2>&1
   else
      exec >"$Log"
      exec 2>&1
   fi
fi

msg "Started $ThisScript v$Version as $USER@${HOSTNAME%%.*} on $(date) as:\\n$0 $*"

# Standard Preflight checks after logging has started.

# Load SDP controlled shell environment.
# shellcheck disable=SC1091
source /p4/common/bin/p4_vars "$SDPInstance" ||\
   errmsg "Failed to load SDP environment."

# shellcheck disable=SC1090
source "$P4CBIN/backup_functions.sh" ||\
   errmsg "Failed to load SDP backup_functions.sh lib.\n"

set_vars

cd "$P4CBIN" || errmsg "Failed to cd to $P4CBIN."
msg "Operating in: $PWD"

msg "Running verify_sdp.sh as part of preflight chekcs."
# Call verify_sdp.sh with honorNoOp=0 so it is called even in NoOp mode.
if run "$P4CBIN/verify_sdp.sh $SDPInstance -skip crontab -L off -online" \
   "Verifying SDP instance $SDPInstance:" 0; then
   msg "\\nVerified: verify_sdp.sh was OK."
else
   errmsg "The verify_sdp.sh script reported preflight errors. Correct these and try again."
fi

for Exe in $ExeList; do
   ExePath="$ExesDir/$Exe"
   if [[ -e "$ExePath" ]]; then
      if [[ -x "$ExePath" ]]; then
         if [[ "$Exe" == "p4" ]]; then
            InstanceLinkPath="${Exe}_bin"
         else
            InstanceLinkPath="${Exe}_${SDPInstance}_bin"
         fi

         if [[ -L "$InstanceLinkPath" ]]; then
            OldMajorVersion=$("./$InstanceLinkPath" -V | grep -i Rev. | awk -F / '{print $3}')
            OldBuildCL=$("./$InstanceLinkPath" -V | grep -i Rev. | awk -F / '{print $4}' | awk '{print $1}')
            OldVersion="${OldMajorVersion}.${OldBuildCL}"
         else
            msg "Skipping $Exe upgrade as there is no current symlink: $InstanceLinkPath"
            continue
         fi

         NewMajorVersion=$("$ExePath" -V | grep -i Rev. | awk -F / '{print $3}')
         NewBuildCL=$("$ExePath" -V | grep -i Rev. | awk -F / '{print $4}' | awk '{print $1}')
         NewVersion="${NewMajorVersion}.${NewBuildCL}"

         if [[ "$OldVersion" != "$NewVersion" ]]; then
            # A lexigraphical comparison is good for our needs here.
            if [[ "$OldVersion" < "$NewVersion" ]]; then
               msg "The $Exe exectuable will be upgraded from $OldVersion to $NewVersion."
               UpgradeExeList+="$Exe "
               [[ "$Exe" != "p4" ]] && UpgradeServiceList+="$Exe "
            else
               errmsg "The staged $Exe exectuable in $ExesDir is an older version [$OldVersion] than the live version [$NewVersion]. Downgrades are not supported."
               continue
            fi
         else
            msg "The $Exe v$NewVersion exectuable does not need to be upgraded."
            continue
         fi

         if [[ "$Exe" == "p4d" ]]; then
            if [[ "$OldMajorVersion" != "$NewMajorVersion" ]]; then
               msg "A major version upgrade for p4d is needed. A schema upgrade will be performed."
               DoP4DMajorVersionUpgrade=1
            else
               msg "P4D patch upgrade is needed, but not a major version upgrade. No schema upgrade will be performed."
            fi

            # A lexigraphical comparison is good for our needs here.
            # shellcheck disable=SC2072
            if [[ "$NewMajorVersion" > "18.2" ]]; then
               msg "P4D version is 2019.1+, so modern upgrade logic will be applied."
               P4D2019Plus=1
            else
               msg "P4D version pre-2019.1, so legacy upgrade logic will be applied."
            fi
         fi

      else
         errmsg "The staged $ExePath is not executable."
      fi
   else
      msg "There is no staged $ExePath binary; skipping $Exe."
   fi
done

if [[ "$ErrorCount" -eq 0 ]]; then
   msg "Verified: Preflight checks passed ($WarningCount warnings)."
else
   bail "Aborting due to failed preflight checks."
fi

if [[ -z "$UpgradeExeList" ]]; then
   msg "No upgrades are needed based on exes staged in $ExesDir."
fi

if [[ "$PreflightOnly" -eq 1 ]]; then
   msg "\\nExiting early after successful preflight checks due to '-p'."
   exit 0
fi

if [[ -z "$UpgradeExeList" ]]; then
   msg "\\nExiting because no upgrades are needed."
   exit 0
fi

#------------------------------------------------------------------------------
# Preflight checks completed. Continue on!

msg "Start $P4SERVER Upgrade as $USER@${HOSTNAME%%.*} at $(date)."

#------------------------------------------------------------------------------
# Phase 1 - Establish a clean rollback point.
if [[ "$UpgradeExeList" == *"p4d"* ]]; then
   if [[ "$SERVERID" == "$P4MASTER_ID" ]]; then
      msg "\\nPhase 1: Establish a clean rollback point for p4d upgrades."
      run "$P4CBIN/rotate_journal.sh $SDPInstance" \
         "Rotating journal to provide a clean rollback point." ||\
         bail "Journal rotation failed. No upgrade processing has been done. Aborting upgrade."

      if grep -q "End $P4SERVER journal rotation" "$LOGS/checkpoint.log"; then
         msg "Verified: Journal rotation completed OK."
      else
         bail "Could not verify journal rotation completed OK. No upgrade processing has been done. Aborting upgrade."
      fi
   else
      msg "Skipping Phase 1 (p4d journal rotation) on replica as this is only needed on master server."
   fi
else
   msg "Skipping Phase 1 (p4d journal rotation) as p4d does not need to be upgraded."
fi

#------------------------------------------------------------------------------
# Phase 2 - Install binaries and update SDP symlinks.
msg "\\nPhase 2: Install binaries and update SDP symlinks in $PWD."

for Exe in $UpgradeExeList; do
   ExePath="$ExesDir/$Exe"
   NewMajorVersion=$("$ExePath" -V | grep -i Rev. | awk -F / '{print $3}')
   NewBuildCL=$("$ExePath" -V | grep -i Rev. | awk -F / '{print $4}' | awk '{print $1}')
   NewVersion="${NewMajorVersion}.${NewBuildCL}"
   VersionLinkPath="${Exe}_${NewMajorVersion}_bin"
   if [[ "$Exe" == "p4" ]]; then
      InstanceLinkPath="${Exe}_bin"
   else
      InstanceLinkPath="${Exe}_${SDPInstance}_bin"
   fi

   OldMajorVersion=$("./$InstanceLinkPath" -V | grep -i Rev. | awk -F / '{print $3}')
   OldBuildCL=$("./$InstanceLinkPath" -V | grep -i Rev. | awk -F / '{print $4}' | awk '{print $1}')
   OldVersion="${OldMajorVersion}.${OldBuildCL}"

   NewVersionNamedExe="${Exe}_$NewVersion"
   if [[ ! -f "$NewVersionNamedExe" ]]; then
      run "cp $ExesDir/$Exe $NewVersionNamedExe" ||\
         errmsg "Failed to copy $ExesDir/$Exe to $NewVersionNamedExe"
   fi
   run "ln -f -s $NewVersionNamedExe $VersionLinkPath" \
      "Updating major version symlink $VersionLinkPath" ||\
      errmsg "Failed to updated major version symlink."
   run "ln -f -s $VersionLinkPath $InstanceLinkPath" \
      "Updating instance symlink $InstanceLinkPath" ||\
      errmsg "Failed to updated major version symlink."
done

#------------------------------------------------------------------------------
# Phase 3 - Stop services to be ugpraed.
if [[ -n "$UpgradeServiceList" ]]; then
   msg "\\nPhase 3: Stop services to be upgraded (if operating): $UpgradeServiceList"

   # Stop services to be upgraded if they have an init script. In the case of a
   # broker, also check that it has a default configuration file, as brokers are
   # often configured but run only on demand with a non-default configuration file
   # (e.g. in maintenance mode with a 'dfm' (Down for Maintenance) config.
   for Exe in $UpgradeExeList; do
      if [[ "$Exe" == "p4broker" ]]; then
         # P4BrokerInitScript is defined with set_vars() from backup_functions.sh.
         # shellcheck disable=SC2154
         if [[ -x "$P4BrokerInitScript" && \
            -f "/p4/common/config/p4_${SDP_INSTANCE}.broker.cfg" ]]; then
            if [[ "$NoOp" -eq 0 ]]; then
               stop_p4broker
            else
               msg "NoOp: Would stop p4broker now."
            fi
         fi
      fi

      if [[ "$Exe" == "p4d" ]]; then
         # P4DInitScript is defined with set_vars() from backup_functions.sh.
         # shellcheck disable=SC2154
         if [[ -x "$P4DInitScript" ]]; then
            if [[ "$NoOp" -eq 0 ]]; then
               stop_p4d
            else
               msg "NoOp: Would stop p4d now."
            fi
         fi
      fi

      if [[ "$Exe" == "p4p" ]]; then
         # P4ProxyInitScript is defined with set_vars() from backup_functions.sh.
         # shellcheck disable=SC2154
         if [[ -x "$P4ProxyInitScript" ]]; then
            if [[ "$NoOp" -eq 0 ]]; then
               stop_p4p
            else
               msg "NoOp: Would stop p4p now."
            fi
         fi
      fi
   done
else
   msg "Skipping Phase 3 (stop services) as only the client needs to be upgraded."
fi

#------------------------------------------------------------------------------
# Phase 4 - Perforce p4d schema upgrades
if [[ "$UpgradeExeList" == *"p4d"* ]]; then
   if [[ "$DoP4DMajorVersionUpgrade" -eq 1 ]]; then
      msg "\\nPhase 4: Perform p4d schema upgrades."

      # For P4D versions >= 19.1, schema upgrades are journaled, and thus are
      # applied to the offline_db through the journal rotation process.
      # For older versions of p4d, we run the 'p4d -xu' separately in the offline_db.
      if [[ "$P4D2019Plus" -eq 1 ]]; then
         UpgradeJournal="$P4JOURNAL"
      else
         UpgradeJournal=off
      fi

      run "$P4DBIN -r $P4ROOT -J $UpgradeJournal -xu" \
         "Upgrading databases in $P4ROOT." ||\
         bail "Failed to upgrade root database."

      # Upgrade the offline database on the master and on edge servers.
      if [[ "$P4REPLICA" == "FALSE" ]] || [[ "$EDGESERVER" -eq 1 ]]; then
         if [[ "$P4D2019Plus" -eq 0 ]]; then
            msg "Upgrading databases in $OFFLINE_DB as this is a pre-19.1 server upgrade."

            # The value of '-t localhost:0000' is necessary for syntax reasons;
            # it is ignored. This allows upgrade of offline databases without needing
            # a license file; there isn't one in offline_db, only P4ROOT.  A valid
            # license file is required to actually start the server process and have
            # it listen on a port.

            run "$P4DBIN -r $OFFLINE_DB -J $UpgradeJournal -xu -t localhost:0000" \
               "Upgrading databases in $OFFLINE_DB." ||\
               bail "Failed to upgrade offline_db"
         fi
      fi
   else
      msg "Skipping Phase 4 (schema upgrade) as p4d upgrade is a patch rather than a major version upgrade." 
   fi
else
   msg "Skipping Phase 4 (schema upgrade) as p4d does not need to be upgraded."
fi

#------------------------------------------------------------------------------
# Phase 5 - Start upgraded services.
if [[ -n "$UpgradeServiceList" ]]; then
   msg "\\nPhase 5: Restart upgraded services: $UpgradeServiceList"

   # Start upgraded services if they have an init script. In the case of a
   # broker, also check that it has a default configuration file, as brokers are
   # often configured but run only on demand with a non-default configuration file
   # (e.g. in maintenance mode with a 'dfm' (Down for Maintenance) config.
   for Exe in $UpgradeExeList; do
      if [[ "$Exe" == "p4broker" ]]; then
         if [[ -x "$P4BrokerInitScript" && \
            -f "/p4/common/config/p4_${SDP_INSTANCE}.broker.cfg" ]]; then
            if [[ "$NoOp" -eq 0 ]]; then
               start_p4broker
            else
               msg "NoOp: Would start p4broker now."
            fi
         fi
      fi

      if [[ "$Exe" == "p4d" ]]; then
         if [[ -x "$P4DInitScript" ]]; then
            if [[ "$NoOp" -eq 0 ]]; then
               start_p4d
            else
               msg "NoOp: Would start p4d now."
            fi
         fi

         # For p4d 2019.1+ upgrades, do an extra journal rotation post-upgrade to apply journal changes to
         # the offline_db.
         if [[ "$P4D2019Plus" -eq 1 ]]; then
            if [[ "$P4REPLICA" == "FALSE" ]] && [[ "$EDGESERVER" -eq 0 ]]; then
               run "$P4CBIN/rotate_journal.sh $SDPInstance" \
                  "Upgrade offline_db via journal rotation on master/commit server." ||\
                  bail "Journal rotation failed during offline_db upgrade phase."
            fi
         fi
      fi

      if [[ "$Exe" == "p4p" ]]; then
         if [[ -x "$P4ProxyInitScript" ]]; then
            if [[ "$NoOp" -eq 0 ]]; then
               start_p4p
            else
               msg "NoOp: Would start p4p now."
            fi
         fi
      fi
   done
else
   msg "Skipping Phase 5 (start services) as only the client needs to be upgraded."
fi

if [[ "$DoProtectionsCommentStyleUpgrade" -eq 1 ]]; then
   if [[ "$P4REPLICA" == "FALSE" && "$EDGESERVER" -eq 0 ]]; then
      msg "Upgrading protections table comment format."
      "$P4BIN" protect --convert-p4admin-comments -o | "$P4BIN" -s protect -i
   fi
fi

msg "Finished $P4SERVER Upgrade."

# See the terminate() function, which is really where this script exits.
exit 0
